/*
 * Copyright (C) 2003-2004 Imendio AB
 * Copyright (C) 2008 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 *
 * Authors: Imendio AB
 *	    Carl Worth <cworth@cworth.org>
 */

#include "loudgame.h"
 
#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
 
void
loudgame_quit (loudgame_t *lg, int return_value)
{
    GError *error = NULL;

    lg->return_value = return_value;

    if (lg->connection)
    {
	if (! lm_connection_close (lg->connection, &error))
	{
	    g_print ("An error occurred during lm_connection_close: %s\n",
		     error->message);
	}
	lm_connection_unref (lg->connection);
	lg->connection = NULL;
    }

    g_main_loop_quit (lg->main_loop);
}
 
static void
authentication_cb (LmConnection *connection, gboolean result, gpointer closure)
{
    LmMessage *m;
    loudgame_t *lg = closure;

    if (! result) {
	g_print ("Authentication for %s failed\n", lg->name);
	loudgame_quit (lg, 1);
	return;
    }
                 
    m = lm_message_new_with_sub_type (NULL,
				      LM_MESSAGE_TYPE_PRESENCE,
				      LM_MESSAGE_SUB_TYPE_AVAILABLE);

    lm_connection_send (connection, m, NULL);
    lm_message_unref (m);
}
 
static void
connection_open_cb (LmConnection *connection, gboolean result, loudgame_t *lg)
{
    lm_connection_authenticate (connection,
				lg->name, lg->passwd, "TestLM",
				authentication_cb, lg, FALSE,  NULL);
}

void
loudgame_send (loudgame_t	*lg,
	       const char	*peer,
	       const char	*message)
{
    LmMessage *reply;
    gboolean result;
    GError *error = NULL;

    reply = lm_message_new (peer, LM_MESSAGE_TYPE_MESSAGE);

    lm_message_node_add_child (reply->node, "body", message);

    result = lm_connection_send (lg->connection, reply, &error);
    lm_message_unref (reply);

    if (! result) {
	g_error ("lm_connection_send failed: %s\n",
		 error->message);
	loudgame_quit (lg, 1);
    }
}

void
loudgame_vsendf (loudgame_t	*lg,
		 const char	*peer,
		 const char	*format,
		 va_list	 va)
{
    char *str;

    str = g_strdup_vprintf (format, va);

    loudgame_send (lg, peer, str);

    free (str);
}

void
loudgame_sendf (loudgame_t	*lg,
		const char	*peer,
		const char	*format,
		...)
{
    va_list va;

    va_start (va, format);

    loudgame_vsendf (lg, peer, format, va);

    va_end (va);
}

typedef struct _loudgame_broadcast_data
{
    loudgame_t *lg;
    const char *message;
} loudgame_broadcast_data_t;

static void
loudgame_broadcast_cb (void *key,
		       void *data,
		       void *closure)
{
    loudgame_broadcast_data_t *lbd = closure;
    char *peer = key;

    loudgame_send (lbd->lg, peer, lbd->message);
}

void
loudgame_broadcast (loudgame_t *lg,
		    const char *message)
{
    loudgame_broadcast_data_t lbd;

    lbd.lg = lg;
    lbd.message = message;

    g_hash_table_foreach (lg->players,
			  loudgame_broadcast_cb, &lbd);
}

void
loudgame_vbroadcastf (loudgame_t *lg,
		      const char *format,
		      va_list     va)
{
    char *str;

    str = g_strdup_vprintf (format, va);

    loudgame_broadcast (lg, str);

    free (str);
}

void
loudgame_broadcastf (loudgame_t	*lg,
		     const char	*format,
		     ...)
{
    va_list va;

    va_start (va,format);

    loudgame_vbroadcastf (lg, format, va);

    va_end (va);
}

static void
handle_command (loudgame_t	*lg,
		const char	*peer,
		const char	*command)
{
    if (strcmp (command, "quit") == 0) {
	loudgame_quit (lg, 0);
	return;
    }

    loudgame_sendf (lg, "Unknown command: '%s'", command);
}

static void
handle_say (loudgame_t	*lg,
	    const char	*peer,
	    const char	*message)
{
    loudgame_broadcastf (lg, "%s: %s", peer, message);
}

static void
handle_whisper (loudgame_t  *lg,
		const char  *peer,
		const char  *body)
{
    const char *recipient, *space, *message;

    space = strchr (body, ' ');
    if (space == NULL) {
	loudgame_sendf (lg, peer,
			"Error: whisper should be of the form 'user message'");
	return;
    }

    recipient = g_strndup (body, space - body);

    message = space + 1;

    loudgame_sendf (lg, recipient, "*%s*: %s", peer, message);
}
 
static LmHandlerResult
handle_messages (LmMessageHandler *handler,
                 LmConnection     *connection,
                 LmMessage        *m,
                 gpointer          closure)
{
    loudgame_t *lg = closure;
    LmMessageNode *body;
    const char *peer;
    const char *body_str;

    peer = lm_message_node_get_attribute (m->node, "from");

    body = lm_message_node_get_child (m->node, "body");
    if (! body)
	return LM_HANDLER_RESULT_REMOVE_MESSAGE;

    if (! g_hash_table_lookup_extended (lg->players, peer,
					NULL, NULL))
    {
	char *peer_copy = g_strdup (peer);
	g_hash_table_insert (lg->players, peer_copy, NULL);
	loudgame_broadcastf (lg, "%s has joined the game", peer);
    }

    body_str = lm_message_node_get_value (body);

    if (body_str) {
	if (body_str[0] == '%')
	    handle_command (lg, peer, body_str + 1);
	else if (strncmp (body_str, "say ", 4) == 0)
	    handle_say (lg, peer, body_str + 4);
	else if (strncmp (body_str, "whisper ", 8) == 0)
	    handle_whisper (lg, peer, body_str + 8);
	else if (lg->handle_message)
	(lg->handle_message) (lg, peer, body_str);
    }
	
    return LM_HANDLER_RESULT_REMOVE_MESSAGE;
}

static gboolean
make_connection (gpointer closure)
{
    loudgame_t	     *lg;
    LmMessageHandler *handler;
    gchar            *jid;
    GError	     *error = NULL;

    lg = closure;
                                                                                
    lg->connection = lm_connection_new (lg->server);

    jid = g_strdup_printf ("%s@%s", lg->name, lg->server);
    lm_connection_set_jid (lg->connection, jid);
    g_free (jid);

    handler = lm_message_handler_new (handle_messages, lg, NULL);
    lm_connection_register_message_handler (lg->connection,
					    handler,
					    LM_MESSAGE_TYPE_MESSAGE,
					    LM_HANDLER_PRIORITY_NORMAL);
                                                                                
    lm_message_handler_unref (handler);
                                                                                
    if (! lm_connection_open (lg->connection,
			      (LmResultFunction) connection_open_cb,
			      lg, NULL, &error))
    {
	g_print ("Failed to open connection to %s: %s.\n",
		 lg->server, error->message);
	loudgame_quit (lg, 1);
    }

    /* It seems to be necessary to explicitly tell the server we're
     * still here. Let's see if one keep-alive every 60 seconds is
     * sufficient. */
    lm_connection_set_keep_alive_rate (lg->connection, 60);

    /* Return false to not schedule another call. */
    return 0;
}

int
loudgame_init (loudgame_t *lg, int argc, char **argv)
{
    if (argc != 4) {
	g_print ("Usage: %s <server> <username> <password>\n", argv[0]);
	return 1;
    }

    lg->server = argv[1];
    lg->name = argv[2];
    lg->passwd = argv[3];

    lg->connection = NULL;

    lg->main_loop = g_main_loop_new (NULL, FALSE);
    lg->players = g_hash_table_new_full (g_str_hash, g_str_equal,
					 g_free, /* key_destroy */
					 NULL    /*value_destroy */);

    lg->return_value = 0;

    lg->handle_message = NULL;

    return 0;
}

static void
loudgame_fini (loudgame_t *lg)
{
    g_main_loop_unref (lg->main_loop);
    g_hash_table_destroy (lg->players);
}

int
loudgame_run (loudgame_t *lg)
{
    g_idle_add (make_connection, lg);

    g_main_loop_run (lg->main_loop);

    loudgame_fini (lg);
                                                                                
    return lg->return_value;
}
